package icfp.contest;

import java.awt.*;
import java.util.*;

/**
 * Insert the type's description here.
 * Creation date: (8/31/02 12:11:12 PM)
 * @author: Pablo J. Pedemonte
 */
public final class Board {

	public final static int POS_INF = Integer.MAX_VALUE;

	public final static char NORTH = 'N';
	public final static char SOUTH = 'S';
	public final static char EAST  = 'E';
	public final static char WEST  = 'W';

	public final static byte VOID  = (byte)'.';
	public final static byte WALL  = (byte)'#';
	public final static byte WATER = (byte)'~';
	public final static byte HOME  = (byte)'@';
	public final static byte BOUND = (byte)'*';
	public final static byte ROBOT = (byte)'R';

	protected final static char[] DIR_ARRAY = new char[] {
		NORTH,EAST,SOUTH,WEST};

	/** Poisition of las visited HOME cell */
	private Point last_home;

	
	private int width;
	private int height;
	private byte[][] board;
/**
 * Board constructor comment.
 */
public Board() {
	super();
}
/**
 * Board constructor comment.
 */
public Board(int height, int width, byte[][] board) {
	super();
	this.height = height;
	this.width  = width;
	this.board  = board;
	this.last_home = new Point();
}
/**
 * Return the ocntesnts of the board at the given cell.
 * If we are outside the limits, return WALL.
 */
public byte cellAt(Point pos) {
	int x = pos.x;
	int y = pos.y;
	if (x < 0 || x >= width || y < 0 || y >= height) {
		return BOUND;
	}

	return board[y][x];
}
/**
 * Insert the method's description here.
 * 
 * @return icfp.contest.NeighbourInfo
 * @param pos java.awt.Point
 * @param dir char
 * @param pn_players java.util.Hashtable
 */
public void fillInNeighbourInfo(
	Point pos, 
	char dir, 
	Hashtable pn_players,
	NeighbourInfo n) {

	n.clear();
		
	byte dist1[] = new byte[4];
	byte dist2[] = new byte[4];
	
	char d;
	Point p1,p2;
	Point q = this.moveInDirection(pos, dir);
	
	// Get info at distance 1 and 2
	q = this.moveInDirection(pos, dir);
	for (int i = 0; i < DIR_ARRAY.length; i++) {
		p1 = this.moveInDirection(pos, DIR_ARRAY[i], false);
		if (pn_players.containsKey(p1)) {
			dist1[i] = ROBOT;
			n.setDist1Occupied(true);
			if (DIR_ARRAY[i] == dir) {
				n.setAimOccupied(true);
			}
		}
		else
			dist1[i] = this.cellAt(p1);
		
		if (q.equals(pos)) {
			continue;
		}
				
		p2 = this.moveInDirection(q, DIR_ARRAY[i], false);
		if (!p2.equals(pos) && pn_players.containsKey(p2)) {
			dist2[i] = ROBOT;
			n.setDist2Occupied(true);
		}
		else
			dist2[i] = this.cellAt(p2);
	}

	n.setDir(dir);
	n.setPos(new Point(pos.x, pos.y));
	n.setDist1(dist1);
	n.setDist2(dist2);
	
	return;
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 7:04:20 PM)
 */
public HashSet findHomeCells() {

	HashSet homeCells = new HashSet();
	
	for(int i = 0; i < this.getHeight(); i++) {
		for (int j = 0; j < this.getWidth(); j++) {
			if (board[i][j] == HOME) {
				homeCells.add(new Point(j,i));
			}
		}
	}

	return homeCells;
}
/**
 * Return the distance to the fill cell where
 * we can get again in the original direction.
 */
private int firstCellFreeAt(Point pos, char origDir, char newDir) {	
	
	int d = 0;
	Point q = new Point(pos.x, pos.y);
	Point r = this.moveInDirection(q, newDir);
	while (!r.equals(q)) {
		++d;
		byte c = this.cellAt(this.moveInDirection(r, origDir, false));
		if (c == BOUND) {
			return POS_INF;
		}
		else if (this.isPassable(c)) {
			return d;
		}

		q = r;
		r = this.moveInDirection(q, newDir);
	}

	return POS_INF;
}
/**
 * Return an alternative direction to origDir for
 * movign from orig to dest, st the returned direction
 * minimizes the traveling cost.
 */
private char getAltDir(Point orig, Point dest, char origDir) {

	int i;
	int c = 0;
	int j = 0;
	int cost = -1;
	int result = 0;
	Point p;
	
	for (i = 0; i < DIR_ARRAY.length; i++) {

		// Ignore original
		if (DIR_ARRAY[i] == origDir) {
			j = i;
			continue;
		}

		// If new direction is not passable, ignore it
		p = this.moveInDirection(orig,DIR_ARRAY[i], true);
		if (!this.isPassable(this.cellAt(p))) {
			c = POS_INF;
		}
		else {
			c = this.getCost(orig, dest, origDir, DIR_ARRAY[i]);
		}

		// Try as much as possible not to undo previous walk
		if (this.oppositeDir(origDir) == DIR_ARRAY[i]) {
			c = Math.max(c * c, c);
		}

		// Get minimum cost
		if (cost < 0 || c < cost) {
			cost = c;
			result = i;
		}
	}

	// if the minimum cost is POS_INF, return the
	// opposite to the original direction
	if (cost == POS_INF) {
		return DIR_ARRAY[(j + 2) % DIR_ARRAY.length];
	}

	return DIR_ARRAY[result];
}
/**
 * Get a heuristic cost for going form orig to dest
 * when changing from the original direction to the
 * new one.
 */
private int getCost(Point orig, Point dest, char origDir, char newDir) {
	int d = this.firstCellFreeAt(orig, origDir, newDir);
	
	if (d == POS_INF) {
		return d;
	}
	
	return d * this.getManhattanDist(orig, dest);
}
/**
 * Get the cheapest direction for moving to orig to
 * dest according to a (hopefully) reasonable heuristic.
 */
public char getDir(Point orig, Point dest) {

	Point p;
	int d = 0;
	int dist = POS_INF;
	
	int minD   = 0;
	Point minP = null;
	
	for (int i = 0; i < DIR_ARRAY.length; i++) {
		p = this.moveInDirection(orig, DIR_ARRAY[i]);
		if (!p.equals(last_home) && this.cellAt(p) == HOME) {
			minD = i;
			minP = p;
			last_home = p;
			break;
		}
		d = this.getManhattanDist(p, dest);
		if (d < dist) {
			dist = d;
			minD = i;
			minP = p;
		}
	}

	byte c = this.cellAt(minP);
	if (this.isPassable(c)) {
		return DIR_ARRAY[minD];
	}

	return this.getAltDir(orig, dest, DIR_ARRAY[minD]);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:12:17 PM)
 * @return int
 */
public int getHeight() {
	return height;
}
/**
 * Get the Manhattan distance from orig to dest.
 */
public int getManhattanDist(Point orig, Point dest) {
	return Math.abs(dest.x - orig.x) + Math.abs(dest.y - orig.y);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:12:17 PM)
 * @return int
 */
public int getWidth() {
	return width;
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 4:51:12 PM)
 * @return boolean
 * @param c byte
 */
protected boolean isPassable(byte c) {
	return c == VOID || c == HOME;
}
/**
 * Return the new position after moving
 * form the origin in the ginven direction.
 *
 * <p>If we fall outside the board, return the
 * original position.
 */
private Point moveInDirection(Point orig, char dir) {
	return this.moveInDirection(orig, dir, true);
}
/**
 * Return the new position after moving
 * form the origin in the ginven direction.
 *
 * <p>If we fall outside the board, and checkBounds 
 * is <code>true</code>, return the original position.
 */
private Point moveInDirection(Point orig, char dir, boolean checkBounds) {
	int x = orig.x;
	int y = orig.y;
	
	switch(dir) {
		case NORTH: ++y; break;
		case SOUTH: --y; break;
		case EAST : ++x; break;
		case WEST : --x; break;
	}

	if (!checkBounds) {
		return new Point(x,y);
	}
	
	if (x < 0) {
	  x = 0;
	}
	else if (x == width) {
	  x = width - 1;
	}

	if (y < 0) {
	  y = 0;
	}
	else if (y == height) {
	  y = height - 1;
	}

	return new Point(x,y);
}
/**
 * Return the opposite direction to the given one
 */
private char oppositeDir(char dir) {
	char opp = '\0';
	switch(dir) {
		case NORTH: opp = SOUTH; break;
		case SOUTH: opp = NORTH; break;
		case EAST : opp = WEST;  break;
		case WEST : opp = EAST;  break;
	}

	return opp;
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:12:17 PM)
 * @param newHeight int
 */
public void setHeight(int newHeight) {
	height = newHeight;
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:12:17 PM)
 * @param newWidth int
 */
public void setWidth(int newWidth) {
	width = newWidth;
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:39:21 PM)
 * @return java.lang.String
 */
public String toString() {
	StringBuffer b = new StringBuffer();
	b.append('(').append(width).append(',').append(height).append(")\n");
	for (int i = 0; i < height; i++) {
		b.append(new String(board[i]));
		b.append('\n');
	}

	return b.toString();
}
}
