package icfp.contest;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A simple game server simulator.
 * 
 * @author: Pablo J. Pedemonte
 */
public class GameServer {

	private String[] b;
	private ServerSocket socket;
/**
 * GameServer constructor comment.
 */
public GameServer(int port, String boardFile) {
	super();
	try {
		socket = new ServerSocket(port);
		this.initialize(boardFile);
		this.start();
	}
	catch(IOException e) {
		System.err.println("Can't start server at port: " + port);
		System.exit(1);
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:17:49 PM)
 * @param fileName java.lang.String
 */
private void initialize(String boardFile) {
	try {
		BufferedReader r = new BufferedReader(new InputStreamReader(
			new FileInputStream(boardFile))
		);

		int row = 0;
		String line = null;

		line = r.readLine();
		if (line == null) {
			System.err.println("Malformed board file. Aborting!");
			return;
		}

		StringTokenizer t = new StringTokenizer(line, " ");
		int height = Integer.parseInt(t.nextToken());
		b = new String[height];
		while((line = r.readLine()) != null) {
			b[row++] = line;
		}
	}
	catch(IOException e) {
		System.err.println("Board file not found: " + boardFile 
			+ ". Aborting!");
		System.exit(1);
	}
}
/**
 * Process this client.
 */
private void processClient(Socket client) {

	String in = null;
	String msg = null;

	try {
		BufferedReader reader = new BufferedReader(
			new InputStreamReader(client.getInputStream())
		);
		BufferedReader stdinReader = new BufferedReader(
			new InputStreamReader(System.in)
		);
		BufferedWriter writer = new BufferedWriter(
			new OutputStreamWriter(client.getOutputStream())
		);

		
		msg = this.recvMessage(reader);
		if (msg.equals("Player")) {
			this.sendInitData(writer);
		}
		else {
			System.err.println("Error: <Player> message expected. Aborting!");
			System.exit(1);
		}
		
		while(true) {
			
			System.out.print("Enter the message to send: ");
			in = stdinReader.readLine();
			if (in == null || in.equalsIgnoreCase("end")) {
				this.terminate();
			}
		    this.sendMessage(writer, in);

		    System.out.print("Enter the message to send: ");
			in = stdinReader.readLine();
			if (msg == null || msg.equalsIgnoreCase("end")) {
				this.terminate();
			}
		    this.sendMessage(writer, in);

			msg = this.recvMessage(reader);
			System.out.println("Received: " + msg);
		}
	}
	catch(IOException e) {
		System.err.println("Unable to process client");
		e.printStackTrace();
		System.exit(1);
	}
}
/**
 * Receive a message from the server.
 * 
 * @return the received message
 */
private String recvMessage(BufferedReader reader) {
	try {
		String msg = reader.readLine();
		if (msg == null) {
			this.terminate();
		}
		return msg;
	}
	catch(IOException e) {
		this.terminate();
		return null;
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:29:15 PM)
 */
private void sendInitData(BufferedWriter writer) {
	
	String dim = b[0].length() + " " + b.length;
	this.sendMessage(writer, dim);
	
	for (int i = 0; i < b.length; i++) {
		this.sendMessage(writer, b[i]);
	}

	String config = "3 25 1000";
	this.sendMessage(writer, config);
}
/**
 * Send amessage to the game server.
 *
 * @param message the message
 */
private void sendMessage(BufferedWriter writer, String message) {
	try {
		writer.write(message.concat("\n"));
		writer.flush();
	}
	catch(IOException e) {
		this.terminate();
	}
}
/**
 * Start the server.
 */
public void start() {
	try {
		final Socket client = socket.accept();
		new Thread(new Runnable() {
			public void run() {
				processClient(client);
			}
		}, "Client_Thread").start();
	}
	catch(IOException e) {
		System.err.println("Unable to bind server socket");
		System.exit(1);
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 12:32:58 PM)
 */
private void terminate() {
	System.err.println("Client down. Shutting down server.");
	System.exit(1);
}
}
