package icfp.contest;

import java.io.*;
import java.awt.*;
import java.net.*;
import java.util.*;

/**
 * A player.
 * 
 * @author: Pablo J. Pedemonte
 */
public class Player {

	private final static int EXIT_SUCCESS = 0;
	private final static int EXIT_FAILURE = 1;
	
	/** Player's socket and streams */
	private Socket socket;
	private BufferedReader reader;
	private BufferedWriter writer;

	/** Player's state */
	private PlayerState state;
/**
 * Player constructor comment.
 */
public Player(String host, int port) {
	super();
	try {
		socket = new Socket(host, port);
		this.initialize();
		this.play();
	}
	catch(IOException e) {
		System.err.println(
			"Player: Can't establish connction with server at: " +
			host + ":" + port);
		System.exit(EXIT_FAILURE);
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 11:34:09 AM)
 */
private void initialize() {
	try {
		reader = new BufferedReader(
			new InputStreamReader(socket.getInputStream()));
		writer = new BufferedWriter(
			new OutputStreamWriter(socket.getOutputStream()));
	}
	catch(IOException e) {
		System.err.println("Error initializing player's connection");
		System.exit(EXIT_FAILURE);
	}
}
/**
 * Player loop.
 */
private void play() {

	new Thread(new Runnable() {
		public void run() {
			startPlaying();
		
			while (true) {
				processPackageUpdates();
				sendCommand();
				processRobotUpdates();
				//System.out.println(state);
			}
		}
	},"Player_Thread").start();
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 11:38:50 AM)
 */
private void processPackageUpdates() {
	String pkgInfo = this.recvMessage();
	StringTokenizer t = new StringTokenizer(pkgInfo, " \t");

	while(t.hasMoreTokens()) {
		Package p = new Package();
		p.uid     = t.nextToken();
		p.pos     = new Point(
						Integer.parseInt(t.nextToken()) - 1, 
						Integer.parseInt(t.nextToken()) - 1);
		p.weight  = Integer.parseInt(t.nextToken());
		state.addPackageToBoard(p);
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 11:56:46 AM)
 */
private void processRobotUpdates() {
	String updateInfo = this.recvMessage();
	state.processRobotUpdates(updateInfo);
}
/**
 * Receive a message from the server.
 * 
 * @return the received message
 */
private String recvMessage() {
	try {
		String msg = reader.readLine();
		if (msg == null) {
			this.stopPlaying();
		}
		return msg;
	}
	catch(IOException e) {
		this.stopPlaying();
		return null;
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 11:52:33 AM)
 */
private void sendCommand() {
	this.sendMessage(state.computeNextAction());
}
/**
 * Send amessage to the game server.
 *
 * @param message the message
 */
private void sendMessage(String message) {
	try {
		writer.write(message.concat("\n"));
		writer.flush();
	}
	catch(IOException e) {
		this.stopPlaying();
	}
}
/**
 * Handshaking before the game starts.
 */
private void startPlaying() {

	String row;
	
	this.sendMessage("Player");
	
	String dim = this.recvMessage();
	StringTokenizer t = new StringTokenizer(dim," \t");
	int width  = Integer.parseInt(t.nextToken());
	int height = Integer.parseInt(t.nextToken());
	byte[][] b = new byte[height][width];
	for(int i = 0; i < height; i++) {
		row = this.recvMessage();
		b[i] = row.getBytes();
	}
	
	Board board = new Board(height, width, b);
	PlayerConfig config = PlayerConfig.parseFrom(this.recvMessage());
	String updateInfo = this.recvMessage();
	state = new PlayerState(board, config);
	state.processRobotUpdates(updateInfo);
}
/**
 * Shutdown routine.
 */
private void stopPlaying() {
	System.err.println("Connection lost with: " + socket 
		+ ". Game finished?");
	System.exit(EXIT_SUCCESS);
}

/**
 * Insert the method's description here.
 * 
 * @param argv java.lang.String[]
 */
public static void main(String[] argv) {
	if (argv.length != 2) {
		System.err.print("Error: invalid usage. Invoke with: ");
		System.err.println("Player <host> <port>");
		System.exit(EXIT_FAILURE);
	}


	int port = 0;
	String host = argv[0];
	try {
		port = Integer.parseInt(argv[1]);
	}
	catch(NumberFormatException e) {
		System.err.println("Error: invalid port number: " + argv[1]);
		System.exit(EXIT_FAILURE);
	}

	// If everything's OK, start playing
	new Player(host, port);
}
}
